
//-----------------------
// Libraries
//-----------------------
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>

#include "taus_rand.h"
#include "matrix.h"

//-----------------------
// Read constants
//-----------------------
#define EPOCH 300
#define LEARNING_RATE 0.001f/BATCH_SIZE //0.0078125f //<-- 96% in 20 epoch for BATCH_SIZE=2

//void printDiff(float *data1, float *data2, int width, int height, float epsilon);
//void printDiff(int *data1, int *data2, int width, int height);
template <int BATCH_SIZE, int VIS_SIZE, int HID_SIZE>
void NeuralNetwork<BATCH_SIZE, VIS_SIZE, HID_SIZE>::solve(unsigned int *data){
    int i, j;
    double result[10];
    double total;

    /*
    FILE *f_weight_file;
    f_weight_file = fopen("weight_seq_file.bin", "rb");
    fread(weight, sizeof(int), VIS_SIZE*HID_SIZE, f_weight_file);
    fclose(f_weight_file);
    */
    
    total = 0;
    for (i = 0; i < 10; i++)
    {
        int visible_l[VIS_SIZE];

        for (j=0; j<VIS_SIZE; j++)
            visible_l[j] = data[j];
        visible_l[486+i] = 1;

        //for (j=0; j<512; j++) printf("%d ", visible_l[j]);
        float energy_l[VIS_SIZE];
        matrix_multS(energy_l, visible_l, weight, 1, VIS_SIZE, HID_SIZE);

        result[i] = 0;        
        for (j = 0; j < HID_SIZE; j++)
            result[i] = result[i] - log(exp(energy_l[j]) + 1);

        printf("New %d=====\n", i);

        printf("result %f, %g\n", result[i], exp(result[i]));
            
        total += exp(-result[i]);
        printf("total %f\n", total);
        printf("\n");
    }
    
    for (i = 0; i < 10; i++)
    {
        printf("Number: %d = %0.3f\n", i, exp(-result[i])/total);
    }
}

template <int BATCH_SIZE, int VIS_SIZE, int HID_SIZE>
NeuralNetwork<BATCH_SIZE, VIS_SIZE, HID_SIZE>::NeuralNetwork(char* wfname, char* trainfname){
    FILE *f_weight_file;
    FILE *f_data_file;
    
    //Read weight file
    f_weight_file = fopen(wfname, "rb");
    fread(weight, sizeof(float), VIS_SIZE*HID_SIZE, f_weight_file);
    fclose(f_weight_file);

    //Read data file
    f_data_file = fopen(trainfname, "rb");
    fread(data, sizeof(int), IMG_SIZE*DATA, f_data_file);
    fclose(f_data_file);
    
    //Device memory
    CUDA_SAFE_CALL( cudaMalloc((void **)&dVisible,   BATCH_SIZE*VIS_SIZE * sizeof(int)));
    CUDA_SAFE_CALL( cudaMalloc((void **)&dVisible_t, VIS_SIZE *BATCH_SIZE* sizeof(int)));
    CUDA_SAFE_CALL( cudaMalloc((void **)&dHidden,    BATCH_SIZE*HID_SIZE * sizeof(int)));

    CUDA_SAFE_CALL( cudaMalloc((void **)&dWeight,   VIS_SIZE*HID_SIZE * sizeof(float)));
    CUDA_SAFE_CALL( cudaMalloc((void **)&dWeight_t, VIS_SIZE*HID_SIZE * sizeof(float)));
    CUDA_SAFE_CALL( cudaMalloc((void **)&dWud_pos,  VIS_SIZE*HID_SIZE * sizeof(int)));
    CUDA_SAFE_CALL( cudaMalloc((void **)&dWud_neg,  VIS_SIZE*HID_SIZE * sizeof(int)));
    
    CUDA_SAFE_CALL( cudaMalloc((void **)&dHidden_energy,  VIS_SIZE*BATCH_SIZE * sizeof(float)));
    CUDA_SAFE_CALL( cudaMalloc((void **)&dVisible_energy, HID_SIZE*BATCH_SIZE * sizeof(float)));
    
    CUDA_SAFE_CALL( cudaMalloc((void **)&dData, IMG_SIZE*DATA * sizeof(float)));

    //Initialize RNG
#ifdef _DEBUG    
    set_seed(0xd0f12aa1, 0xac4b0b42, 0x5aef557d);
#else
    initializeRNG();
#endif

}

template <int BATCH_SIZE, int VIS_SIZE, int HID_SIZE>
template <int WEIGHT_WIDTH>
void NeuralNetwork<BATCH_SIZE, VIS_SIZE, HID_SIZE>::transfer(int* from, int* to, float* from_energy, float* weights){

    dim3 threads(BLOCK_SIZE, BLOCK_SIZE);	
    dim3 grid(BATCH_SIZE/threads.x, WEIGHT_WIDTH/threads.y);
    matrixMult<BATCH_SIZE, WEIGHT_WIDTH, int, float> <<< grid, threads >>> (from_energy, from, weights);
	
#ifdef _DEBUG    
    CUDA_SAFE_CALL( cudaMemcpy(visible_energy, from_energy, HID_SIZE*BATCH_SIZE* sizeof(int), cudaMemcpyDeviceToHost));
    
    for (int i = 0; i < HID_SIZE*BATCH_SIZE; i++)
    {
        float random=(((float)get_urand())/0xFFFFFFFF);
        visible[i] = (1/(1 + exp(-visible_energy[i])) > random);
    }
    visible[0] = 1;
    CUDA_SAFE_CALL( cudaMemcpy(to, visible, HID_SIZE*BATCH_SIZE* sizeof(int), cudaMemcpyHostToDevice) );
#else
    GPUtransfer<<<8, 128>>>(to, from_energy, N_PER_RNG);
#endif
}

template <int BATCH_SIZE, int VIS_SIZE, int HID_SIZE>
void NeuralNetwork<BATCH_SIZE, VIS_SIZE, HID_SIZE>::train(){
    CUDA_SAFE_CALL( cudaMemcpy(dData, data, IMG_SIZE*DATA* sizeof(int), cudaMemcpyHostToDevice) );
    CUDA_SAFE_CALL( cudaMemcpy(dWeight, weight, VIS_SIZE*HID_SIZE* sizeof(float), cudaMemcpyHostToDevice) );  
    dim3 threads(BLOCK_SIZE, BLOCK_SIZE);
	
    for (int m = 0; m < EPOCH; m++) {
        for (int i = 0; i < BATCH_SIZE*VIS_SIZE; i++)
            visible[i] = data[i];
		
        CUDA_SAFE_CALL( cudaMemcpy(dVisible, visible, BATCH_SIZE*VIS_SIZE* sizeof(int), cudaMemcpyHostToDevice) );
          
        transfer<HID_SIZE>(dVisible, dHidden, dVisible_energy, dWeight);

        matrix_transpose<BATCH_SIZE, VIS_SIZE, int>(dVisible_t, dVisible);
		
    	dim3 grid1(BATCH_SIZE/threads.x, VIS_SIZE/threads.y);
        matrixMult<VIS_SIZE, HID_SIZE, int, int> <<< grid1, threads >>> (dWud_pos, dVisible_t, dHidden);
      	
      	matrix_transpose<VIS_SIZE, HID_SIZE, float>(dWeight_t, dWeight);
		
        transfer<VIS_SIZE>(dHidden, dVisible, dHidden_energy, dWeight_t); 
        transfer<HID_SIZE>(dVisible, dHidden, dVisible_energy, dWeight);

        CUDA_SAFE_CALL( cudaMemcpy(visible, dVisible, BATCH_SIZE*VIS_SIZE*sizeof(int), cudaMemcpyDeviceToHost));
		
        matrix_transpose<BATCH_SIZE, VIS_SIZE, int>(dVisible_t, dVisible);

        matrixMult<VIS_SIZE, HID_SIZE, int, int> <<< grid1, threads >>> (dWud_neg, dVisible_t, dHidden);
        
//#ifdef _DEBUG
        CUDA_SAFE_CALL( cudaMemcpy(visible, dVisible, VIS_SIZE*BATCH_SIZE, cudaMemcpyDeviceToHost));
        recon_valid = 0;
        int label_recon = 0;
        for (int i = 0; i < VIS_SIZE*BATCH_SIZE; i++)
            if (visible[i] == data[i])
                recon_valid++;

        for (int i = 0; i < BATCH_SIZE; i++)
            for (int j = 0; j < 10; j++)
                 if (visible[i*VIS_SIZE + j + 486] == data[i*VIS_SIZE + j + 486])
                    label_recon++;
        printf("%8d, %f, %f\n", m, (float)recon_valid/(VIS_SIZE*BATCH_SIZE), (float)label_recon/(10*VIS_SIZE));
//#endif        
        dim3 grid(VIS_SIZE/threads.x, HID_SIZE/threads.y);
        matrixMac<VIS_SIZE, HID_SIZE, int> <<< grid, threads >>>(dWeight, dWud_pos, dWud_neg, LEARNING_RATE);
    }

    CUDA_SAFE_CALL( cudaMemcpy(weight, dWeight, VIS_SIZE*HID_SIZE*sizeof(float), cudaMemcpyDeviceToHost));

    CUDA_SAFE_CALL(cudaFree(dVisible));
    CUDA_SAFE_CALL(cudaFree(dVisible_t));
    CUDA_SAFE_CALL(cudaFree(dHidden));
    CUDA_SAFE_CALL(cudaFree(dWeight));
    CUDA_SAFE_CALL(cudaFree(dWeight_t));
    CUDA_SAFE_CALL(cudaFree(dWud_pos));
    CUDA_SAFE_CALL(cudaFree(dWud_neg));
    CUDA_SAFE_CALL(cudaFree(dHidden_energy));
    CUDA_SAFE_CALL(cudaFree(dVisible_energy));
    CUDA_SAFE_CALL(cudaFree(dData)); 
}

#ifdef _DEBUG
template <>
void printDeviceArray<float>(float* arr, int size, float* loc_temp){ 
    CUDA_SAFE_CALL( cudaMemcpy(loc_temp, arr, size*sizeof(float), cudaMemcpyDeviceToHost));
    printf("\n===========================================================\n");
    CUDA_SAFE_CALL(cudaThreadSynchronize());
    for (int i = 0; i < size && i < 200; i++) //&& i < 200
        printf("%f ", loc_temp[i]);
}

template <>
void printDeviceArray<int>(int* arr, int size, int* loc_temp){ 
    CUDA_SAFE_CALL( cudaMemcpy(loc_temp, arr, size*sizeof(int), cudaMemcpyDeviceToHost));
    printf("\n===========================================================\n");
    CUDA_SAFE_CALL(cudaThreadSynchronize());
    for (int i = 0; i < size && i < 200; i++) //&& i < 200
        printf("%d ", loc_temp[i]);
}  
#endif

void printDiff(float *data1, float *data2, int width, int height, float epsilon)
{
  int i,j,k;
  int error_count=0;
  for (j=0; j<height; j++) {
    for (i=0; i<width; i++) {
      k = j*width+i;
      if (abs(data1[k] - data2[k]) > epsilon) {
         printf("diff(%d,%d) CPU=%4.4f, GPU=%4.4f n", i,j, data1[k], data2[k]);
         error_count++;
      }
    }
  }
  printf(" nTotal Errors = %d n", error_count);
}

void printDiff(int *data1, int *data2, int width, int height)
{
  int i,j,k;
  int error_count=0;
  for (j=0; j<height; j++) {
    for (i=0; i<width; i++) {
      k = j*width+i;
      if (data1[k] != data2[k]) {
         printf("diff(%d,%d) CPU=%4.4f, GPU=%4.4f n", i,j, data1[k], data2[k]);
         error_count++;
      }
    }
  }
  printf(" nTotal Errors = %d n", error_count);
}

