#ifndef __NEURAL_NETWORK_H__
#define __NEURAL_NETWORK_H__

#define IMG_SIZE 512
#define DATA 1024
//#define _DEBUG

#ifdef _DEBUG
template <typename T>
void printDeviceArray(T* arr, int size, T* loc_temp);
#endif

template <int BATCH_SIZE, int VIS_SIZE, int HID_SIZE>
class NeuralNetwork {
  public:
    NeuralNetwork(char* wfname, char* trainfname);
    
    void train();
    void solve(unsigned int *data);
  
  private:    
    template <int WEIGHT_WIDTH>
    void inline transfer(int* from, int* to, float* from_energy, float* weights);
    
    // Internal state variables
#ifdef _DEBUG //Pull data off the device for debugging purposes
    
    int visible_t[VIS_SIZE*BATCH_SIZE];
    int hidden   [BATCH_SIZE*HID_SIZE];
    
    float weight_t[HID_SIZE*VIS_SIZE];
    int wud_pos [VIS_SIZE*HID_SIZE];
    int wud_neg [VIS_SIZE*HID_SIZE];
    
    float hidden_energy [BATCH_SIZE*VIS_SIZE];
    float visible_energy[BATCH_SIZE*HID_SIZE];
    
    int tempdata  [BATCH_SIZE*VIS_SIZE];
#endif

    int recon_valid;
    int visible  [BATCH_SIZE*VIS_SIZE];
    
    float weight[VIS_SIZE*HID_SIZE];
    int   data  [IMG_SIZE*DATA]; //training data
    
    //Device memory: most of the internal state of the 
    int* dVisible;
    int* dVisible_t;    
    int* dHidden;
    //int* tempdVisible_t;

    float* dWeight;
    float* dWeight_t;
    int* dWud_pos;
    int* dWud_neg;
    
    float* dHidden_energy;
    float* dVisible_energy;
    
    int* dData;
};

#include "NeuralNetwork.cpp"
#endif
