#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <cutil.h>
#include <math.h>
#include "MersenneTwister.h"
#include "MersenneTwister_kernel.cu"

extern "C"{
    //ceil(a / b)
    int iDivUp(int a, int b){
        return ((a % b) != 0) ? (a / b + 1) : (a / b);
    }

    //floor(a / b)
    int iDivDown(int a, int b){
        return a / b;
    }

    //Align a to nearest higher multiple of b
    int iAlignUp(int a, int b){
        return ((a % b) != 0) ?  (a - a % b + b) : a;
    }

    //Align a to nearest lower multiple of b
    int iAlignDown(int a, int b){
        return a - a % b;
    }

    void initMTRef(const char *fname);
}

void initializeRNG(char **dir); 


// Data configuration
const int    PATH_N    = 24000000;
const int    N_PER_RNG = iAlignUp(iDivUp(PATH_N, MT_RNG_COUNT), 2);
const int    RAND_N    = MT_RNG_COUNT * N_PER_RNG;

const unsigned int SEED = 777; //change this part to - time(NULL)


int main(int argc, char **argv){
	//declare variables here

	CUT_DEVICE_INIT(argc, argv);
    CUT_SAFE_CALL( cutCreateTimer(&hTimer) );
	
	//Initialize data and memory allocation here
	
	initializeRNG(argv);
}

//Executable and main .cu file must have the same name!
void initializeRNG(char **dir) {
	printf("Loading CPU and GPU twisters configurations...\n");
        const char *raw_path = cutFindFilePath("MersenneTwister.raw", dir[0]);
        const char *dat_path = cutFindFilePath("MersenneTwister.dat", dir[0]);
		//printf("path: %s\n", raw_path);
        initMTRef(raw_path);
        loadMTGPU(dat_path);
        seedMTGPU(SEED);
}
