#ifndef _MATRIX_H_
#define _MATRIX_H_

// Thread block size
#define BLOCK_SIZE 16

// Matrix dimensions
// (chosen as multiples of the thread block size for simplicity)
#define WA (16 * BLOCK_SIZE) // Matrix A width
#define HA (16 * BLOCK_SIZE) // Matrix A height
#define WB (16 * BLOCK_SIZE) // Matrix B width
#define HB WA  // Matrix B height
#define WC WB  // Matrix C width 
#define HC HA  // Matrix C height

void matrix_multS(float *C, float *A, float *B, int hA, int wA, int wB);
void matrix_transposeS(float *B, float *A, int hA, int wA);
void matrix_mac(float *C, float *A, float *B, int hA, int wA, float epsilon);
//void matrix_test();

////////////////////////////////////////////////////////////////////////////////
// Compute matrix multiply
////////////////////////////////////////////////////////////////////////////////

void matrix_multS(float *C, float *A, float *B, int hA, int wA, int wB)
{
    int i, j, k;
    float sum;

    for (i = 0; i < hA; i++)
    {
        for (j = 0; j < wB; j++)
        {
            sum = 0;
            for (k = 0; k < wA; k++)
                sum += A[i*wA + k]*B[k*wB + j];

            C[i*wB + j] = sum;
        }
    }
}

void matrix_multS(float *C, int *A, float *B, int hA, int wA, int wB)
{
    int i, j, k;
    float sum;

    for (i = 0; i < hA; i++)
    {
        for (j = 0; j < wB; j++)
        {
            sum = 0;
            for (k = 0; k < wA; k++)
                sum += A[i*wA + k]*B[k*wB + j];

            C[i*wB + j] = sum;
        }
    }
}

void matrix_multS(int *C, int *A, int *B, int hA, int wA, int wB)
{
    int i, j, k;
    int sum;

    for (i = 0; i < hA; i++)
    {
        for (j = 0; j < wB; j++)
        {
            sum = 0;
            for (k = 0; k < wA; k++)
                sum += A[i*wA + k]*B[k*wB + j];

            C[i*wB + j] = sum;
        }
    }
}

////////////////////////////////////////////////////////////////////////////////
// Compute matrix tranpose
////////////////////////////////////////////////////////////////////////////////

void matrix_transposeS(float *B, float *A, int hA, int wA)
{
    int i, j;

    for (i = 0; i < hA; i++)
        for (j = 0; j < wA; j++)
            B[j*hA + i] = A[i*wA + j];
}

void matrix_transposeS(int *B, int *A, int hA, int wA)
{
    int i, j;

    for (i = 0; i < hA; i++)
        for (j = 0; j < wA; j++)
            B[j*hA + i] = A[i*wA + j];
}

////////////////////////////////////////////////////////////////////////////////
// Compute matrix mac
////////////////////////////////////////////////////////////////////////////////

void matrix_mac(float *C, float *A, float *B, int hA, int wA, float epsilon)
{
    int i, j;

    for (i = 0; i < hA; i++)
        for (j = 0; j < wA; j++)
            C[i*wA + j] += epsilon*(A[i*wA + j] - B[i*wA + j]);
}

void matrix_mac(float *C, int *A, int *B, int hA, int wA, float epsilon)
{
    int i, j;

    for (i = 0; i < hA; i++)
        for (j = 0; j < wA; j++)
            C[i*wA + j] += epsilon*(A[i*wA + j] - B[i*wA + j]);
}

////////////////////////////////////////////////////////////////////////////////
// Quick test
////////////////////////////////////////////////////////////////////////////////
/*
void matrix_test()
{
    int i, j;
    float A[4][5];
    float AT[5][4];
    float B[5][3];
    float C[4][3];

    for (i = 0; i < 4; i++)
        for (j = 0; j < 5; j++)
            A[i][j] = i+j;

    for (i = 0; i < 4; i++)
    {
        for (j = 0; j < 5; j++)
            printf("%d ", A[i][j]);
        printf("\n");
    }
    printf("\n");

    matrix_transpose(&AT[0][0], &A[0][0], 4, 5);
    for (i = 0; i < 5; i++)
    {
        for (j = 0; j < 4; j++)
            printf("%d ", AT[i][j]);
        printf("\n");
    }
    printf("\n");

    for (i = 0; i < 5; i++)
        for (j = 0; j < 3; j++)
            B[i][j] = 5*(i+j+1);

    for (i = 0; i < 5; i++)
    {
        for (j = 0; j < 3; j++)
            printf("%d ", B[i][j]);
        printf("\n");
    }
    printf("\n");

    matrix_mult(&C[0][0], &A[0][0], &B[0][0], 4, 5, 3);

    for (i = 0; i < 4; i++)
    {
        for (j = 0; j < 3; j++)
            printf("%d ", C[i][j]);
        printf("\n");
    }
    printf("\n");
}
*/
#endif 

