/*         1
 *  o = --------
 *       1 +  1
 *           ---
 *           e^x
 */


/*
 * Simplest function, just do the actual operation using **FLOATS**
 */
__device__ float sigmoidF (float g_idata){
    return 1/(1+__expf(-g_idata));
}

#define SCALE
#define CHAR_E 2.71828182845904523536


/*            SCALE
 *  o = -----------------
 *       SCALE  +     SCALE
 *                -------------
 *               (E_SCALE^X_SCALE)
 *
 */

/*
 * Sigmoid curve split into 
 */
__device__ float sigmoidACu (float g_idata){
    int m = 32; float b = 1.0f; //in>8
    signed int in = (signed int)(g_idata*134217728); //
    // out = in<<m + b, find m and b
    if      (in<-8<<27) { m = 32; b = 0.0000f; } 
    else if (in<-4<<27) { m =  6; b = 0.1250f; }
    else if (in<-2<<27) { m =  5; b = 0.1875f; } 
    else if (in<-1<<27) { m =  3; b = 0.3750f; } 
    else if (in< 1<<27) { m =  2; b = 0.5000f; }
    else if (in< 2<<27) { m =  3; b = 0.6250f; }
    else if (in< 4<<27) { m =  5; b = 0.8125f; } 
    else if (in< 8<<27) { m =  6; b = 0.8750f; }
    return (float)(in>>m)/134217728 + b;
}

/*
 * TEXTURE implementation
 */
#define TEX_SIZE 1024 
#define RANGE 10.0f // +/- RANGE
 
// declare texture reference for 1D float texture
texture<float, 1, cudaReadModeElementType> tex;

void setupTexture() { 
    printf("Initializing texture.\n");
    //Generate texture values
    float x=-RANGE;
    float *h_txdata = (float*) malloc(TEX_SIZE*sizeof(float));
    for (int i=0; i<TEX_SIZE; i++, x+=(2*RANGE/TEX_SIZE))
        h_txdata[i] = 1/(1+expf(-x));
    
    //for (int i=0; i<TEX_SIZE; i++)
    //    printf("%f", h_txdata[i]);
    
    printf("\n");    
    // allocate array and copy image data
    cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc(32, 0, 0, 0, cudaChannelFormatKindFloat);
    cudaArray* cu_array;
    CUDA_SAFE_CALL( cudaMallocArray  (&cu_array, &channelDesc,   TEX_SIZE*sizeof(float), 1)); 
    CUDA_SAFE_CALL( cudaMemcpyToArray( cu_array, 0, 0, h_txdata, TEX_SIZE*sizeof(float), cudaMemcpyHostToDevice));

    // set texture parameters
    tex.addressMode[0] = cudaAddressModeClamp;
    tex.filterMode     = cudaFilterModeLinear;
    tex.normalized     = false;
    tex.channelDesc    = channelDesc;

    // Bind the array to the texture
    CUDA_SAFE_CALL( cudaBindTextureToArray(tex, cu_array, channelDesc));
}

__device__ float sigmoidTX (float g_idata){
    return tex1D(tex, g_idata*TEX_SIZE/RANGE + TEX_SIZE/2); 
    //Address calculation should use multiply and add (a single instruction) if compiler did constant folding
}

/*
 * Second order aproximation with two pieces
 */
__device__ float sigmoid2ndOrd (float g_idata){
    float in = g_idata;
    if      (in<-4.f) in=-4.f;
    else if (in> 4.f) in= 4.f;
    float partial = (1-in*in)/2;
    if (in<0) partial = 1 - partial;
    return partial;
}
